package CardGames.cards;

import java.util.*;

/**
 * This class extends SuitedCard to handle cards from a standard 52-card deck with four suits,
 * and card values from 1 (Ace) to 13 (King).  Implements Comparable so that hands can be sorted
 * in card value order easily by hand-counting algorithms.
 *
 * @author Mark Fitzgerald
 * @version 2002.Jan.03
 */
public class StandardCard extends SuitedCard
{
    /**
     * Default constructor.  Creates an ace of clubs.
     */
    public StandardCard()
    {
        super(1,1);
    }

    /**
     * Constructor for a standard card.
     *
     * @param value The value of the card.  Valid values are between 1 (ace) and 13 (king), inclusive.  (12 is queen, 11 is jack.)
     * @param suit The suit of the card.  Valid values are between 1 (clubs) and 4 (spades), inclusive.  (2 is diamonds, 3 is hearts.)
     */
    public StandardCard(int value, int suit) throws InvalidCard
    {
        super(value, suit);
        validate();
    }

    /** The exception type thrown for invalid card creation attempts. */
    public static class InvalidCard extends Exception
    {
        public InvalidCard() {}
        public InvalidCard(String msg) { super(msg); }
    }
    
    /**
     * Validate the value and suit of this card.  Used in both constructors.
     */
    private void validate() throws InvalidCard
    {
        if (mValue < 1 || mValue > 13)
        {
            throw new InvalidCard("Invalid card value.");
        }
        else if (mSuit < 1 || mSuit > 4)
        {
            throw new InvalidCard("Invalid card suit.");
        }
    }

    /**
     * Constructor for a standard card.
     *
     * @param strCard The string representation of the card.  Consists of 1 or 2 characters of value (A, 2, ..., 10, J, Q, K) followed by 1 character of suit (CHSD).
     */
    public StandardCard(String strCard) throws InvalidCard
    {
        super(0,0);
        try
        {
            mValue = Integer.parseInt(strCard.substring(0,strCard.length()-1));
            if (mValue < 2 || mValue > 10)
            {
                throw new InvalidCard("Please specify cards outside the 2-9 range non-numerically.");
            }
        }
        catch (NumberFormatException nfEx)
        {
            //Try to parse it as a 1-character AJQK.
            switch (Character.toUpperCase(strCard.charAt(0)))
            {
                case 'A':
                    mValue = 1;
                    break;
                case 'J':
                    mValue = 11;
                    break;
                case 'Q':
                    mValue = 12;
                    break;
                case 'K':
                    mValue = 13;
                    break;
                default:
                    throw new InvalidCard("Invalid standard card value.");
            }
        }

        switch (Character.toUpperCase(strCard.charAt(strCard.length()-1)))
        {
            case 'S':
                mSuit = 4;
                break;
            case 'H':
                mSuit = 3;
                break;
            case 'D':
                mSuit = 2;
                break;
            case 'C':
                mSuit = 1;
                break;
            default:
                throw new InvalidCard("Invalid standard card suit.");
        }
        validate();
    }

    /**
     * Return the string representation of the card, using A, J, Q, K for values, and S, H, C, D for suits.
     * @return A string representation of the card.  The returned string has the form "f_s", where f is the value, and s is the suit.
     */
    public String toString()
    {
        String strValue = null;
        switch (mValue)
        {
            case 1:
                strValue = "A";
                break;
            case 11:
                strValue = "J";
                break;
            case 12:
                strValue = "Q";
                break;
            case 13:
                strValue = "K";
                break;
            default:
                strValue = Integer.toString(mValue);
        }

        String strSuit = null;
        switch (mSuit)
        {
            case 1:
                strSuit = "C";
                break;
            case 2:
                strSuit = "D";
                break;
            case 3:
                strSuit = "H";
                break;
            default:
                strSuit = "S";
        }

        return strValue + strSuit;
    }

    /**
     * Generate a deck of 52 unique cards - a standard deck.
     * @return A <code>Vector</code> containing 52 unique <code>StandardCard</code>s.
     */
    public static Vector generateDeck()
    {
        Vector deck = new Vector(52);
        for (int suit=1; suit<=4; ++suit)
        {
            for (int value=1; value<=13; ++value)
            {
                try
                {
                    deck.add(new StandardCard(value, suit));
                }
                catch (InvalidCard e)
                {
                    //Ignore - no exception should occur here.
                    System.out.println("Bad card generated in generateDeck()?");
                }
            }
        }
        return deck;
    }
}