package CardGames.cribbage;

import CardGames.cards.StandardCard;
import CardGames.probability.CombinationIterator;
import java.util.*;

/** So far, just a tool for calculating the average point values of randomly generated
 * hands.
 *
 * Hope to in the future calculate the average best point value of sets of 5 and 6
 * cards, as well as best value including crib (+, = or -), and best value including
 * crib and pegging (+ or =).
 *
 * @author Mark Fitzgerald
 * @version 2002.Oct.31
 */
public class Trivia
{
    /** Continuously count randomly generated hands, and summarize the average points of the 5 types periodically.
     * @param iNExtraCards The number of cards above 4 to give to the player. Can be 0, 1 or 2.
     * @throws IllegalArgumentException If the number of extra cards is invalid. */
    public static void AverageRandomHandCount(int iNExtraCards)
    {
        CribbageHandCountCounts oAllCounts = new CribbageHandCountCounts();
        StandardCard[] thisHand;
        StandardCard thisTurnup;

        Vector deck = StandardCard.generateDeck();
        final double dCheckpointScaling = 1.4d;
        long lTotalHandsDealt = 0L;
        long lNextCheckpoint = 100L;
        long lHandsToNextCheckpoint = lNextCheckpoint;
        while (true)
        {
            if (iNExtraCards == 0)
            {
                oAllCounts.Add(dealAndCountRandomHand(deck, false));
            }
            else if (iNExtraCards == 1 || iNExtraCards == 2)
            {
                oAllCounts.Add(dealAndCountRandomHand(deck, iNExtraCards, false));
            }
            else
            {
                 throw new IllegalArgumentException("Bad number of extra cards: " + iNExtraCards + ".");
            }
            
            --lHandsToNextCheckpoint;
            if (lHandsToNextCheckpoint == 0)
            {
                lTotalHandsDealt += lNextCheckpoint - lTotalHandsDealt;
                lNextCheckpoint = Math.round(dCheckpointScaling * ((double)lNextCheckpoint));
                lHandsToNextCheckpoint = lNextCheckpoint - lTotalHandsDealt;
                
                System.out.println("Cache statistics: " + Utilities.getCacheStatistics()[0]
                    + " puts to cache, " + Utilities.getCacheStatistics()[1] + " gets from cache.");
                printAverages(lTotalHandsDealt, oAllCounts);
            }
        }
    }
    
    /** Shuffles the deck, generates a random hand from the deck, and returns its
     * score.
     * @return The in-hand points for the random hand.
     * @see Utilities#countHand */
    private static CribbageHandCount dealAndCountRandomHand(Vector deck, boolean bIsCrib)
    {
        Collections.shuffle(deck);  //Takes about 85% of the time in this method.
        return Utilities.countHand(new StandardCard[]{ (StandardCard)deck.get(0),
            (StandardCard)deck.get(1), (StandardCard)deck.get(2),
            (StandardCard)deck.get(3)}, (StandardCard)deck.get(4), bIsCrib);
    }
    
    /** Shuffles the deck, generates a random set of 4-plus cards from the deck, discards <code>iNExtraCards</code>
     * cards, then counts the hand against all <code>48-iNExtraCards</code> possible turnups, totalling the value. Do
     * the same for the other possible discard combinations, and select the hand which has the highest total number of
     * points against all turnups.
     * @param iNExtraCards The number of cards above 4 to give to the player. Can be 0, 1 or 2.
     * @return The points for the best hand summed against all <code>48-iNExtraCards</code> turnups.
     * @throws IllegalArgumentException If the number of extra cards is invalid.
     * @see Utilities#countHand */
    private static CribbageHandCountCounts dealAndCountRandomHand(Vector deck, int iNExtraCards, boolean bIsCrib)
    {
        if (iNExtraCards < 0 || iNExtraCards > 2)
        {
            throw new IllegalArgumentException("Bad number of extra cards: " + iNExtraCards + ".");
        }
        Collections.shuffle(deck);  //Takes about 85% of the time in this method.
        double[] adCount = null;
        CribbageHandCountCounts oBestDiscardCounts = null;
        CombinationIterator objComb = new CombinationIterator(4+iNExtraCards, 4);
        long lBestTotalCount = -1L;
        int iBestComb = -1;
        while (objComb.hasNextCombination())
        {
            int[] aiComb = objComb.nextCombination();
            CribbageHandCountCounts oAllCountsForThisComb = new CribbageHandCountCounts();
            for (int iTurnup = 4+iNExtraCards; iTurnup < 52; ++iTurnup)
            {
                CribbageHandCount oCount = Utilities.countHand(new StandardCard[]{ (StandardCard)deck.get(aiComb[0]),
                    (StandardCard)deck.get(aiComb[1]), (StandardCard)deck.get(aiComb[2]),
                    (StandardCard)deck.get(aiComb[3])}, (StandardCard)deck.get(iTurnup), bIsCrib);
                oAllCountsForThisComb.Add(oCount);
            }
            final long lTotalPoints = oAllCountsForThisComb.GetPoints();
            if (lTotalPoints > lBestTotalCount)
            {
                lBestTotalCount = lTotalPoints;
                oBestDiscardCounts = oAllCountsForThisComb;
            }
        }
        return oBestDiscardCounts;
    }
    
    private static void printAverages(long lTotalHandsDealt, CribbageHandCountCounts oAllScores)
    {
        //All padded to 8 characters for output alignment purposes.
        System.out.println("Average point values after " + lTotalHandsDealt + " hands:");
        System.out.println("Fifteens: " + oAllScores.GetAverageFifteensPoints());
        System.out.println("Pairs   : " + oAllScores.GetAveragePairsPoints());
        System.out.println("Runs    : " + oAllScores.GetAverageRunsPoints());
        System.out.println("Flush   : " + oAllScores.GetAverageFlushPoints());
        System.out.println("His Nobs: " + oAllScores.GetAverageHisNobsPoints());
        System.out.println("Average total points: " + oAllScores.GetAveragePoints());
    }
}