package CardGames.cribbage;

import CardGames.cards.StandardCard;
import CardGames.probability.CombinationIterator;
import java.util.*;

/**
 * This class contains utility functions for <code>SuitedCard</code>s in the game of Cribbage.
 *
 * @author Mark Fitzgerald
 * @version 2002.Jan.05
 */
public class Utilities
{
    /**
     * Determine the counting value of the StandardCard.
     * @param card The <code>StandardCard</code> to count.
     * @return The counting value of <code>card</code>, a value from 1 to 10 inclusive.
     */
    public static int countingValue(StandardCard card)
    {
        return (card.mValue>10)?(10):(card.mValue);
    }

    /**
     * A place for the fifteens+runs+pairs count for a hand to be stored.  The hash indices are formed
     * by sorting the hand in increasing order of card value, then using the formula a*13^4+b*13^3+c*13^2+d*13+e,
     * where a is the first card value and e is the last card value, with values between 0 and 12 (not 1 and 13).
     * I believe 3094 unique keys will exist.  Actual runs indicate >=6175 unique keys, however (perhaps 6188?).
     */
    private static HashMap countHandCache = new HashMap(10000);

    private static int cachePuts = 0;
    private static int cacheGets = 0;
    /**
     * @return Cache statistics in an <code>int[2]</code>, with index 0 being the number of cache puts, and index 1 being the number of cache gets.
     */
    public static int[] getCacheStatistics()
    {
        return new int[]{cachePuts, cacheGets};
    }

    /**
     * Return the point value of the hand described by the array of 4 hand cards and 1 turn-up card.
     * @param hand The 4 cards in hand.  Exception thrown if array not of length 4.
     * @param turnup The turned up card.
     * @param isCrib <code>true</code> if the hand to be counted is the crib.  <code>false</code> otherwise.  If <code>true</code>, 4-card flushes do not count.
     * @return The number of points in-hand.
     * @throws IllegalArgumentException if a hand with other than 4 cards is handed in.
     */
    public static CribbageHandCount countHand(StandardCard[] hand, StandardCard turnup, boolean isCrib)
    {
        if (hand.length != 4)
        {
            throw new IllegalArgumentException(hand.length + " cards were handed in instead of 4.");
        }

        //Check for his Nobs.
        int nobsPoints = 0;
        for (int nobsIndex=0; nobsIndex<4; ++nobsIndex)
        {
            if ((hand[nobsIndex].mValue == 11) && (hand[nobsIndex].mSuit == turnup.mSuit))
            {
                nobsPoints = 1;
                break;
            }
        }

        //Check for a flush.
        int flushPoints = 4;
        for (int flushIndex=0; flushIndex<3; ++flushIndex)
        {
            if (hand[flushIndex].mSuit != hand[flushIndex+1].mSuit)
            {
                flushPoints = 0;
                break;
            }
        }
        if (hand[3].mSuit == turnup.mSuit && flushPoints == 4)
        {
            flushPoints++;
        }
        if (isCrib && flushPoints == 4)
        {
            flushPoints = 0;
        }

        //Created a hand sorted by value to simplify the search for runs (and to allow calculation of a
        //small set of unique hash keys).
        Vector vectFullHand = new Vector(5);
        for (int index=0; index<4; ++index)
        {
            vectFullHand.add(hand[index]);
        }
        vectFullHand.add(turnup);
        Collections.sort(vectFullHand);

        //Check if the runs + fifteens + pairs count is cached.
        Integer hashKey = null;
        int iHashKey = 0;
        for (int handIndex=0; handIndex<5; ++handIndex)
        {
            iHashKey += ((StandardCard)vectFullHand.get(handIndex)).mValue * Math.round(Math.pow(13,handIndex));
        }
        hashKey = new Integer(iHashKey);
        Vector cachedCount = (Vector)(countHandCache.get(hashKey));

        //Now count, whether it is cached or must be calculated (and then cached).
        int runsPoints = 0;
        int pairsPoints = 0;
        int fifteensPoints = 0;
        if (cachedCount != null)
        {
            runsPoints = ((Integer)(cachedCount.get(0))).intValue();
            pairsPoints = ((Integer)(cachedCount.get(1))).intValue();
            fifteensPoints = ((Integer)(cachedCount.get(2))).intValue();
            ++cacheGets;
        }
        else
        {
            //Find all runs.
            for (int nCards=5; nCards>=3 && runsPoints==0; --nCards)
            {
                CombinationIterator subsetIterator = new CombinationIterator(5, nCards);
                while (subsetIterator.hasNextCombination())
                {
                    int[] combination = subsetIterator.nextCombination();
                    boolean isRun = true;
                    for (int combinationIndex=0; combinationIndex<combination.length-1; ++combinationIndex)
                    {
                        StandardCard firstCard = (StandardCard)vectFullHand.get(combination[combinationIndex]);
                        StandardCard secondCard = (StandardCard)vectFullHand.get(combination[combinationIndex+1]);
                        if (firstCard.mValue + 1 != secondCard.mValue)
                        {
                            isRun = false;
                            break;
                        }
                    }
                    if (isRun)
                    {
                        runsPoints += nCards;
                    }
                }
            }

            //A 5-array makes counting pairs and fifteens much easier.
            StandardCard[] fullHand = new StandardCard[5];
            for (int index=0; index<4; ++index)
            {
                fullHand[index] = hand[index];
            }
            fullHand[4] = turnup;

            //Find all pairs.
            CombinationIterator pairIterator = new CombinationIterator(5, 2);
            while (pairIterator.hasNextCombination())
            {
                int[] combination = pairIterator.nextCombination();
                if (fullHand[combination[0]].mValue == fullHand[combination[1]].mValue)
                {
                    pairsPoints += 2;
                }
            }

            //Find all fifteens.
            for (int nCards=2; nCards<=5; ++nCards)
            {
                CombinationIterator subsetIterator = new CombinationIterator(5, nCards);
                while (subsetIterator.hasNextCombination())
                {
                    int[] combination = subsetIterator.nextCombination();
                    int thisTotal = 0;
                    for (int combinationIndex=0; combinationIndex<combination.length; ++combinationIndex)
                    {
                        thisTotal += countingValue(fullHand[combination[combinationIndex]]);
                    }
                    if (thisTotal == 15)
                    {
                        fifteensPoints += 2;
                    }
                }
            }

            //Cache the counts.
            cachedCount = new Vector(3);
            cachedCount.add(new Integer(runsPoints));
            cachedCount.add(new Integer(pairsPoints));
            cachedCount.add(new Integer(fifteensPoints));
            countHandCache.put(hashKey, cachedCount);
            ++cachePuts;
        }

        return new CribbageHandCount(fifteensPoints, pairsPoints, runsPoints, flushPoints, nobsPoints);
    }
}
