package CardGames.probability;

import java.util.*;

/**
 * This class allows one to iterate over all combinations of size R<=N from a set of size N.
 * @author Mark Fitzgerald
 * @version 2002.Jan.04
 */
public class CombinationIterator
{
    /**
     * Create a combination iterator.
     * @param setSize The number of items in the set.  Must be at least 1.
     * @param subsetSize The number of items in each subset (combination) to be returned.  Must be at least 1,
     * and <= <code>setSize</code>.
     * @throws IllegalArgumentException if either parameter is <= 0, or if the subset size is larger than the set size.
     */
    public CombinationIterator (int setSize, int subsetSize)
    {
        if (setSize < 1)
        {
            throw new IllegalArgumentException("Set size must be at least 1.");
        }
        if (subsetSize < 1 || subsetSize > setSize)
        {
            throw new IllegalArgumentException("Subset size must be at least 1, and cannot exceed the "
                + "size of the set, " + setSize + ".");
        }

        miSetSize = setSize;
        miSubsetSize = subsetSize;

        currentCombination = new int[subsetSize];
        for (int index=0; index<subsetSize-1; ++index)
        {
            currentCombination[index] = index;
        }
        currentCombination[subsetSize-1] = subsetSize-2;  //So that the first 'next' call returns the right value.
    }

    /**
     * Return whether another combination exists.
     * @return True if another combination exists, false otherwise.
     */
    public boolean hasNextCombination()
    {
        return !((currentCombination[0] == miSetSize - miSubsetSize)
            && (currentCombination[miSubsetSize-1] == miSetSize - 1));
    }

    /**
     * Return the next combination.
     * @return The next combination if another combination exists, otherwise, an <code>int[0]</code>.
     */
    public int[] nextCombination()
    {
        if (!hasNextCombination())
        {
            return new int[0];
        }

        //Increment the combination.
        incrementCombination();

        //Continue until no duplicate indices exist, and the indices are in ascending order.
        while (noDuplicates() == false || inAscendingOrder() == false)
        {
            incrementCombination();
        }

        return currentCombination;
    }

    /**
     * Increment the combination as if it were a base-<code>miSetSize</code> number.
     */
    private void incrementCombination()
    {
        for (int digit=miSubsetSize-1; digit>=0; --digit)
        {
            currentCombination[digit]++;
            if (currentCombination[digit] < miSetSize)
            {
                break;
            }
            else
            {
                currentCombination[digit] = 0;
            }
        }
    }

    /**
     * State whether any duplicates entries exist in this combination.
     * @return True if no duplicated indices, false otherwise.
     */
    private boolean noDuplicates()
    {
        for (int first=0; first<miSubsetSize-1; ++first)
        {
            for (int second=first+1; second<miSubsetSize; ++second)
            {
                if (currentCombination[first] == currentCombination[second])
                {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * State whether the indices are in ascending order.  We want combinations here, not permutations!  Will do this
     * by returning only the in-order permutations (one exists per each combination).
     * @return True if the indices are in ascending order, false otherwise.
     */
    private boolean inAscendingOrder()
    {
        for (int index=0; index<miSubsetSize-1; ++index)
        {
            if (currentCombination[index] >= currentCombination[index+1])
            {
                return false;
            }
        }
        return true;
    }

    private int miSetSize, miSubsetSize;
    private int[] currentCombination;

    /**
     * This method is simply used for unit-testing the CombinationIterator class.
     * @param args Command-line arguments.  Specify either none, or two, with the first being the size of the set, and the second being the size of subsets to take.
     */
    public static void main(String[] args)
    {
        //Check the arguments.
        int setSize=5, subsetSize=2;
        if (args.length < 2)
        {
            System.out.println("Specify two arguments, the first being the set size, the second being the subset size.");
            System.out.println("Choosing default values of 5 and 2.");
        }
        else
        {
            try
            {
                setSize = Integer.parseInt(args[0]);
                subsetSize = Integer.parseInt(args[1]);
                System.out.println("Parsed the command-line arguments.");
            }
            catch (NumberFormatException exNumberFormat)
            {
                System.out.println("Could not parse the command-line arguments.");
                return;
            }
        }

        //Iterate over all combinations.
        CombinationIterator combIter = null;
        try
        {
            combIter = new CombinationIterator(setSize, subsetSize);
            System.out.println("Created an iterator over all subsets of size " + subsetSize + " on a set of size " + setSize + ".");
        }
        catch (Exception e)
        {
            System.out.println("Could not create the CombinationIterator!");
        }

        //Go through all possible combinations.
        long combinationIndex = 1;
        while (combIter.hasNextCombination())
        {
            System.out.print("Combination " + combinationIndex++ + ": ");
            int[] combination = combIter.nextCombination();
            for (int index=0; index<combination.length-1; ++index)
            {
                System.out.print(combination[index] + ",");
            }
            System.out.println(combination[combination.length-1] + ".");
        }
    }
}
